# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from iconsdk.exception import KeyStoreException, DataTypeException
from iconsdk.utils.hexadecimal import (
    is_0x_prefixed,
    remove_0x_prefix,
    is_cx_prefixed,
    remove_cx_prefix,
    is_hx_prefixed,
    remove_hx_prefix
)
from iconsdk.utils.type import is_str, is_integer


def is_keystore_file(keystore: dict) -> bool:
    """Checks data in a keystore file is valid.

    :return: type(bool)
        True: When format of the keystore is valid.
        False: When format of the keystore is invalid.
    """

    root_keys = ["version", "id", "address", "crypto", "coinType"]
    crypto_keys = ["ciphertext", "cipherparams", "cipher", "kdf", "kdfparams", "mac"]
    crypto_cipherparams_keys = ["iv"]

    is_valid = has_keys(keystore, root_keys) and has_keys(keystore["crypto"], crypto_keys) \
               and has_keys(keystore["crypto"]["cipherparams"], crypto_cipherparams_keys)

    if is_valid:
        return is_valid
    else:
        raise KeyStoreException("The keystore file is invalid.")


def has_keys(target_data: dict, keys: list) -> bool:
    """Checks to a target data for having all of keys in list."""
    for key in keys:
        if key not in target_data.keys():
            return False
    return True


def is_keystore_file_for_icon(keystore: dict) -> bool:
    """
    Checks to a keystore for not eth but icon.
    1. Checks that a value of a key 'address' starts with 'hx'.
    2. Checks that a value of a key 'coinType' is same as 'icx'
    """
    if is_wallet_address(keystore["address"]) and keystore["coinType"] == "icx":
        return True
    else:
        raise KeyStoreException("The keystore file is invalid.")


def is_wallet_address(value) -> bool:
    """
    Checks if value is T_ADDR_EOA type.
    T_ADDR_EOA is data type which is 40-digit hexadecimal string prefixed with `hx`.

    :param value: wallet address
    """
    return is_str(value) and value.islower() and is_hx_prefixed(value) and len(remove_hx_prefix(value)) == 40


def is_score_address(value) -> bool:
    """
    Checks if value is T_ADDR_SCORE type.
    T_ADDR_SCORE is data type which is 40-digit hexadecimal string prefixed with `cx`.

    :param value: SCORE address
    """
    return is_str(value) and value.islower() and is_cx_prefixed(value) and len(remove_cx_prefix(value)) == 40


def is_T_HASH(value) -> bool:
    """T_HASH is data type which is 64-digit hexadecimal string prefixed with `0x`."""
    try:
        if is_0x_prefixed(value) and len(remove_0x_prefix(value)) == 64:
            return True
        else:
            raise DataTypeException("This hash value is unrecognized.")
    except ValueError:
        raise DataTypeException("This hash value is unrecognized.")


def is_T_BIN_DATA(value) -> bool:
    """
    T_BIN_DATA is data type which is hexadecimal string prefixed with `0x`
    and length is even.
    """
    try:
        if is_0x_prefixed(value) and len(remove_0x_prefix(value)) % 2 == 0:
            return True
        else:
            raise DataTypeException("This value is not T_BIN_DATA data type.")
    except ValueError:
        raise DataTypeException("This value is not T_BIN_DATA data type.")


def is_predefined_block_value(value) -> bool:
    """
    By far, predefined block value is only `latest`.
    Later it is possible to add others.

    :param value: "latest". type(str)
    :return: type(bool)
    """
    return is_str(value) and value == "latest"


def is_hex_block_hash(value) -> bool:
    """
    Checks the value - a parameter is valid.
    Hash value of a block starts with '0x' and 64 digits hex string

    :param value: hash value of a block, hexadecimal digits. type(str)
    :return: type(bool)
    """
    return is_str(value) and is_0x_prefixed(value) and len(remove_0x_prefix(value)) == 64


def is_block_height(value: int) -> bool:
    """Checks the value - a parameter is valid.

    :param value: height of a block, hexadecimal digits. type(str).
    :return: type(bool)
    """
    try:
        if not is_integer(value):
            return False
    except ValueError:
        return False
    return 0 <= value < 2 ** 256


def is_block(result: dict) -> bool:
    """Checks block information in response has right format.

    :param result
    :return: bool
    """
    inner_key_of_result = ["version", "prev_block_hash", "merkle_tree_root_hash", "time_stamp",
                           "confirmed_transaction_list", "block_hash", "height", "peer_id", "signature"]
    return has_keys(result, inner_key_of_result)


def is_score_apis(result: list) -> bool:
    """Checks list of score apis in response has right format.

    :param result
    :return: bool
    """
    for api in result:
        if api['type'] == 'eventlog':
            inner_key_of_result = ["type", "name", "inputs"]
        else:
            inner_key_of_result = ["type", "name", "inputs", "outputs"]
        inner_key_of_inputs = ["name", "type"]
        if not has_keys(api, inner_key_of_result):
            return False
        if len(api["inputs"]) > 0 and not has_keys(api["inputs"][0], inner_key_of_inputs):
            return False
    return True


def is_transaction(result: dict) -> bool:
    """Checks the result of `icx_getTransactionByHash` has right format.

    :param result
    :return: bool
    """
    inner_key_of_result = ["version", "from", "to", "stepLimit", "timestamp", "nid", "txIndex",
                           "blockHeight", "blockHash", "signature"]
    return has_keys(result, inner_key_of_result)


def is_transaction_result(result: dict) -> bool:
    """Checks the result of `icx_getTransactionResult` has right format.

    :param result
    :return: bool
    """
    inner_key_of_result = ["status", "to", "txHash", "txIndex", "blockHeight", "blockHash",
                           "cumulativeStepUsed", "stepUsed", "stepPrice"]
    return has_keys(result, inner_key_of_result)


def is_basic_transaction(params: dict) -> bool:
    """
    Checks an instance of `Transaction` has right format.
    Every types of `Transaction` like icx transaction, deploy transaction, call transaction and message transaction
    is checked by this method.
    """
    inner_key_of_params = ['version', 'from', 'to', 'stepLimit', 'timestamp', 'nid']

    return has_keys(params, inner_key_of_params) \
           and is_wallet_address(params['from']) \
           and (is_wallet_address(params['to']) or is_score_address(params['to'])) \
           and is_0x_prefixed(params['stepLimit']) \
           and is_0x_prefixed(params['timestamp']) is not None


def is_icx_transaction(params: dict) -> bool:
    """Checks an instance of `Transaction` for transfer icx has right format."""
    return is_basic_transaction(params) \
           and 'value' in params \
           and is_0x_prefixed(params['value'])


def is_deploy_transaction(params: dict) -> bool:
    """Checks an instance of `DeployTransaction` has right format."""
    inner_key_of_params = ['dataType', 'data']
    inner_key_of_data = ['contentType', 'content']

    return is_basic_transaction(params) \
           and has_keys(params, inner_key_of_params) \
           and has_keys(params['data'], inner_key_of_data) \
           and params['dataType'] == 'deploy' \
           and is_0x_prefixed(params['data']['content']) \
           and 'value' not in params


def is_call_transaction(params: dict) -> bool:
    """Checks an instance of `CallTransaction` has right format."""
    inner_key_of_params = ['dataType', 'data']
    inner_key_of_data = ['method']
    return is_basic_transaction(params) \
           and has_keys(params, inner_key_of_params) \
           and has_keys(params["data"], inner_key_of_data) \
           and params["dataType"] == "call"


def is_message_transaction(params: dict) -> bool:
    """Checks an instance of `MessageTransaction` has right format."""
    inner_key_of_params = ['dataType', 'data']
    return is_basic_transaction(params) \
           and has_keys(params, inner_key_of_params) \
           and is_0x_prefixed(params["data"]) \
           and params["dataType"] == "message"


def is_deposit_transaction(params: dict, deposit_type: str) -> bool:
    """Checks an instance of `DepositTransaction` has right format.

    :param params:
    :param deposit_type: add or withdraw
    :return: bool
    """
    inner_key_of_params = ['dataType', 'data']
    inner_key_of_data_for_add_action = ["action"]
    inner_key_of_data_for_withdraw_action = ["action", "id"]

    checked: bool = has_keys(params, inner_key_of_params) and params["dataType"] == "deposit"

    if deposit_type == "withdraw":
        return has_keys(params["data"], inner_key_of_data_for_withdraw_action) \
               and is_0x_prefixed(params["data"]["id"]) and params["data"]["action"] == "withdraw" and checked
    else:
        return has_keys(params["data"], inner_key_of_data_for_add_action) \
               and params["data"]["action"] == "add" and checked



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Optional, Dict, Any


class MonitorSpec(metaclass=ABCMeta):
    @abstractmethod
    def get_request(self) -> any:
        """
        Request object to send to the websocket
        """
        raise NotImplementedError("MonitorSpec must implement this method")

    @abstractmethod
    def get_path(self) -> str:
        """
        Extra path fragment for the websocket
        """
        raise NotImplementedError("MonitorSpec must implement this method")


class MonitorTimeoutException(Exception):
    pass


class Monitor(metaclass=ABCMeta):
    @abstractmethod
    def close(self):
        """
        Close the monitor

        It releases related resources.
        """
        pass

    @abstractmethod
    def read(self, timeout: Optional[float] = None) -> any:
        """
        Read the notification

        :param timeout: Timeout to wait for the message in fraction of seconds
        :except MonitorTimeoutException: if it passes the timeout
        """
        pass


class Provider(metaclass=ABCMeta):
    """The provider defines how the IconService connects to RPC server."""

    @abstractmethod
    def make_request(self, method: str, params: Optional[Dict[str, Any]] = None, full_response: bool = False):
        raise NotImplementedError("Providers must implement this method")

    @abstractmethod
    def make_monitor(self, spec: MonitorSpec, keep_alive: Optional[float] = None) -> Monitor:
        """
        Make monitor for the spec
        :param spec: Monitoring spec
        :param keep_alive: Keep-alive message interval in fraction of seconds
        """
        raise NotImplementedError()




from collections import namedtuple
from enum import Enum

from iconsdk.utils.convert_type import convert_hex_str_to_bytes, convert_hex_str_to_int


def addVersion(key, value, new_obj):
    new_obj[key] = value if value != '' else 2


def convert_data_content_to_bytes(key, value, new_obj):
    if new_obj["dataType"] == "deploy":
        new_obj[key]["content"] = convert_hex_str_to_bytes(value["content"])


def convert_failure_code_to_int(key, value, new_obj):
    if "failure" in new_obj:
        new_obj[key]["code"] = convert_hex_str_to_int(value["code"])


BLOCK_0_1A_VERSION = '0.1a'
BLOCK_0_3_VERSION = '0.3'

ConvertKeyName = namedtuple("ConvertKeyName", "new_key")
RemoveKey = namedtuple("RemoveKey", "")
ExceptionHandle = namedtuple("ExceptionHandle", "func")

funcAddVersion = addVersion
funcConvertDataContentToBytes = convert_data_content_to_bytes
funcConvertFailureCodeToBytes = convert_failure_code_to_int


class ValueType(Enum):
    none = 0
    str = 1
    int = 2
    bytes = 3
    list = 4
    dict = 5
    hex_hash_number = 6
    prefixed_hex_hash_number = 7


TRANSACTIONS_OF_BLOCK = [
    {
        "CHANGE": {
            "tx_hash": ConvertKeyName("txHash"),
            "method": RemoveKey(),
        },
        "version": ValueType.int,
        "from": ValueType.str,
        "to": ValueType.str,
        "value": ValueType.int,
        "fee": ValueType.int,
        "stepLimit": ValueType.int,
        "timestamp": ValueType.int,
        "nid": ValueType.int,
        "nonce": ValueType.int,
        "signature": ValueType.str,
        "txHash": ValueType.prefixed_hex_hash_number,
        "dataType": ValueType.str,
        "data": ValueType.dict
    }
]

TRANSACTIONS_OF_GENESIS_BLOCK = [
    {
        "accounts": ValueType.list,
        "message": ValueType.str,
        "nid": ValueType.int
    }
]

TRANSACTION = {
    "CHANGE": {
        "tx_hash": ConvertKeyName("txHash"),
        "method": RemoveKey(),
        "version": ExceptionHandle(funcAddVersion),
        "data": ExceptionHandle(funcConvertDataContentToBytes)
    },
    "version": ValueType.int,
    "from": ValueType.str,
    "to": ValueType.str,
    "value": ValueType.int,
    "fee": ValueType.int,
    "stepLimit": ValueType.int,
    "timestamp": ValueType.int,
    "nid": ValueType.int,
    "nonce": ValueType.int,
    "signature": ValueType.str,
    "txHash": ValueType.prefixed_hex_hash_number,
    "dataType": ValueType.str,
    "data": ValueType.dict,
    "txIndex": ValueType.int,
    "blockHeight": ValueType.int,
    "blockHash": ValueType.prefixed_hex_hash_number
}

TRANSACTION_RESULT = {
    "CHANGE": {
        "failure": ExceptionHandle(funcConvertFailureCodeToBytes)
    },
    "status": ValueType.int,
    "to": ValueType.str,
    "txHash": ValueType.prefixed_hex_hash_number,
    "txIndex": ValueType.int,
    "blockHeight": ValueType.int,
    "blockHash": ValueType.prefixed_hex_hash_number,
    "cumulativeStepUsed": ValueType.int,
    "stepUsed": ValueType.int,
    "stepPrice": ValueType.int,
    "eventLogs": ValueType.list,
    "logsBloom": ValueType.bytes,
    "failure": ValueType.dict,
    "scoreAddress": ValueType.str
}

PREV_VOTES = [
    {
        "rep": ValueType.str,
        "timestamp": ValueType.int,
        "blockHeight": ValueType.int,
        "blockHash": ValueType.hex_hash_number,
        "signature": ValueType.str
    }
]

LEADER_VOTES = [
    {
        "rep": ValueType.str,
        "timestamp": ValueType.int,
        "blockHeight": ValueType.int,
        "oldLeader": ValueType.str,
        "newLeader": ValueType.str,
        "signature": ValueType.str
    }
]

BLOCK_0_1a = {
    "version": ValueType.str,
    "prev_block_hash": ValueType.hex_hash_number,
    "merkle_tree_root_hash": ValueType.hex_hash_number,
    "time_stamp": ValueType.int,
    "confirmed_transaction_list": TRANSACTIONS_OF_BLOCK,
    "block_hash": ValueType.hex_hash_number,
    "height": ValueType.int,
    "peer_id": ValueType.str,
    "signature": ValueType.str,
    "next_leader": ValueType.str
}

BLOCK_0_3 = {
    "CHANGE": {
        "prev_block_hash": ConvertKeyName("prevHash"),
        "merkle_tree_root_hash": ConvertKeyName("transactionsHash"),
        "time_stamp": ConvertKeyName("timestamp"),
        "confirmed_transaction_list": ConvertKeyName("transactions"),
        "block_hash": ConvertKeyName("hash"),
        "peer_id": ConvertKeyName("leader"),
        "next_leader": ConvertKeyName("nextLeader"),
    },
    "version": ValueType.str,
    "hash": ValueType.prefixed_hex_hash_number,
    "prevHash": ValueType.prefixed_hex_hash_number,
    "prevVotesHash": ValueType.prefixed_hex_hash_number,
    "transactionsHash": ValueType.prefixed_hex_hash_number,
    "stateHash": ValueType.prefixed_hex_hash_number,
    "receiptsHash": ValueType.prefixed_hex_hash_number,
    "repsHash": ValueType.prefixed_hex_hash_number,
    "nextRepsHash": ValueType.prefixed_hex_hash_number,
    "leaderVotesHash": ValueType.prefixed_hex_hash_number,
    "logsBloom": ValueType.bytes,
    "timestamp": ValueType.int,
    "height": ValueType.int,
    "leader": ValueType.str,
    "nextLeader": ValueType.str,
    "signature": ValueType.str,
    "transactions": TRANSACTIONS_OF_BLOCK,
    "prevVotes": PREV_VOTES,
    "leaderVotes": LEADER_VOTES
}



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import warnings
from abc import ABCMeta, abstractmethod
from hashlib import sha3_256
from typing import Dict, Union, Any

from coincurve import PrivateKey, PublicKey
from eth_keyfile import create_keyfile_json, extract_key_from_keyfile, decode_keyfile_json
from multimethod import multimethod

from iconsdk import logger
from iconsdk.exception import DataTypeException, KeyStoreException
from iconsdk.libs.signer import sign
from iconsdk.utils import PathLikeObject, store_keystore_file_on_the_path
from iconsdk.utils.validation import is_keystore_file


class Wallet(metaclass=ABCMeta):
    """An interface `Wallet` has 2 abstract methods, `get_address()` and `sign(hash: str)`."""

    @abstractmethod
    def get_address(self) -> str:
        """Gets a wallet address of wallet which starts with 'hx'.

        :return address: A wallet address
        """
        raise NotImplementedError("Wallet must implement this method")

    @abstractmethod
    def sign(self, data: bytes) -> bytes:
        """Generates signature from input data which is transaction data

        :param data: data to be signed
        :return signature: signature made from input
        """
        raise NotImplementedError("Wallet implement this method")


class KeyWallet(Wallet):
    """KeyWallet class implements Wallet."""

    def __init__(self, private_key_object: PrivateKey):
        self._private_key_object: PrivateKey = private_key_object

    @property
    def private_key(self) -> bytes:
        return self._private_key_object.secret

    @property
    def public_key(self) -> bytes:
        return self._private_key_object.public_key.format(compressed=False)

    @staticmethod
    def create() -> 'KeyWallet':
        """Generates an instance of Wallet without a specific private key.

        :return: An instance of Wallet class.
        """
        private_key_object = PrivateKey()
        wallet = KeyWallet(private_key_object)
        logger.info(f"Created Wallet. Address: {wallet.get_address()}")
        return wallet

    @multimethod
    def load(private_key: bytes) -> KeyWallet:
        """Loads a wallet from a private key and generates an instance of Wallet.

        :param private_key: private key in bytes
        :return: An instance of Wallet class.
        """
        try:
            wallet = KeyWallet(PrivateKey(private_key))
            logger.info(f"Loaded Wallet by the private key. Address: {wallet.get_address()}")
            return wallet
        except TypeError:
            raise DataTypeException("Private key is invalid.")

    @staticmethod
    @multimethod
    def load(file_path: PathLikeObject, password: str) -> KeyWallet:
        """Loads a wallet from a keystore file with your password and generates an instance of Wallet.

        :param file_path: File path of the keystore file. type(str | bytes | PathLike)
        :param password:
            Password for the keystore file.
            It must include alphabet character, number, and special character.
        :return: An instance of Wallet class.
        """
        try:
            with open(file_path, 'rb') as file:
                private_key: bytes = extract_key_from_keyfile(file, bytes(password, 'utf-8'))
                private_key_object = PrivateKey(private_key)
                wallet = KeyWallet(private_key_object)
                logger.info(
                    f"Loaded Wallet by the keystore file. Address: {wallet.get_address()}, File path: {file_path}")
                return wallet
        except FileNotFoundError as e:
            raise KeyStoreException(f'File not found: {e}')
        except ValueError as e:
            raise KeyStoreException(f'Wrong password: {e}')
        except Exception as e:
            raise KeyStoreException(f'Keystore error: {e}')

    def store(self, file_path: PathLikeObject, password: str):
        """Stores data of an instance of a derived wallet class on the file path with your password.

        :param file_path: File path of the keystore file. type(str | bytes | PathLike)
        :param password:
            Password for the keystore file. Password must include alphabet character, number, and special character.
            type(str)
        """
        try:
            key_store_contents = self.to_dict(password)
            # validate the  contents of a keystore file.
            if is_keystore_file(key_store_contents):
                json_string_keystore_data = json.dumps(key_store_contents)
                store_keystore_file_on_the_path(file_path, json_string_keystore_data)
                logger.info(f"Stored Wallet. Address: {self.get_address()}, File path: {file_path}")
        except FileExistsError:
            raise KeyStoreException("File already exists.")
        except PermissionError:
            raise KeyStoreException("Not enough permission.")
        except FileNotFoundError:
            raise KeyStoreException("File not found.")
        except IsADirectoryError:
            raise KeyStoreException("Directory is invalid.")

    def to_dict(self, password: str) -> Dict[str, Any]:
        ret: Dict[str, Any] = create_keyfile_json(
            self.private_key,
            bytes(password, 'utf-8'),
            iterations=16384,
            kdf="scrypt"
        )
        ret['address'] = self.get_address()
        ret['coinType'] = 'icx'
        return ret

    @classmethod
    def from_dict(cls, jso: Dict[str, Any], password: str) -> KeyWallet:
        private_key: bytes = decode_keyfile_json(jso, password)
        return KeyWallet.load(private_key)

    def get_private_key(self, hexadecimal: bool = True) -> Union[str, bytes]:
        """Returns the private key of the wallet.
        """
        pri_key: bytes = self._private_key_object.secret
        return pri_key.hex() if hexadecimal else pri_key

    def get_public_key(self, compressed: bool = True, hexadecimal: bool = True) -> Union[str, bytes]:
        pub_key: bytes = self._private_key_object.public_key.format(compressed)
        return pub_key.hex() if hexadecimal else pub_key

    def get_address(self) -> str:
        """Returns an EOA address.

        :return address: An EOA address
        """
        return public_key_to_address(self._private_key_object.public_key.format(compressed=False))

    def sign(self, data: bytes) -> bytes:
        """Generates signature from input data which is transaction data

        :param data: data to be signed
        :return signature: signature made from input
        """
        return sign(data, self.private_key)

    def __eq__(self, other: KeyWallet) -> bool:
        return self.private_key == other.private_key

    def __ne__(self, other: KeyWallet) -> bool:
        return not self.__eq__(other)

    def __deepcopy__(self, memodict={}) -> KeyWallet:
        return KeyWallet.load(self.private_key)

    def __hash__(self):
        return hash(self._private_key_object.secret)


def public_key_to_address(public_key: bytes) -> str:
    if not (len(public_key) == 65 and public_key[0] == 4):
        pub_key = PublicKey(public_key)
        public_key: bytes = pub_key.format(compressed=False)
    return f'hx{sha3_256(public_key[1:]).digest()[-20:].hex()}'


def convert_public_key_format(public_key: bytes, compressed: bool) -> bytes:
    pub_key = PublicKey(public_key)
    return pub_key.format(compressed)


def get_public_key(private_key_object: PrivateKey):
    warnings.warn("get_public_key is deprecated, use KeyWallet.public_key", DeprecationWarning)
    return private_key_object.public_key.format(compressed=False)



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import wraps
from logging import Formatter, StreamHandler
from time import time
from typing import Union

from iconsdk import logger

# https://docs.python.org/3/glossary.html#term-path-like-object
PathLikeObject = Union[str, bytes, os.PathLike]


def store_keystore_file_on_the_path(file_path: PathLikeObject, json_string: str):
    """Stores a created keystore string data which is JSON format on the file path.

    :param file_path: The path where the file will be saved. type(str)
    :param json_string: Contents of the keystore.
    """
    if os.path.isfile(file_path):
        raise FileExistsError

    with open(file_path, 'wt') as f:
        f.write(json_string)


def apply_to_return_value(callback):
    def outer(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            return callback(fn(*args, **kwargs))

        return inner

    return outer


to_dict = apply_to_return_value(dict)


def set_logger(level, handler=StreamHandler(),
               format: str = '%(asctime)s %(name)-12s %(levelname)-5s '
                             '%(filename)-12s %(lineno)-4s %(funcName)-12s %(message)s') -> None:
    """ Set logger by setting level and handler

    :param level: the logging level of this logger. The level must be an int or a str.
    :param handler: the specified handler to be added to this logger
    :param format: the specified format strings which initialize the formatter with.
    :return: None
    """
    formatter = Formatter(format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)


def get_timestamp() -> str:
    """Returns the timestamp in microseconds since the epoch"""
    return hex(int(time() * 10 ** 6))



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from iconsdk.exception import DataTypeException
from iconsdk.utils.hexadecimal import add_0x_prefix, remove_0x_prefix
from iconsdk.utils.type import is_integer, is_bytes, is_str


def convert_int_to_hex_str(value: int) -> str:
    """Converts an integer to hex string prefixed with '0x'."""
    if is_integer(value):
        return hex(value)
    else:
        raise DataTypeException("Data type should be integer.")


def convert_bytes_to_hex_str(value: bytes) -> str:
    """Converts bytes to hex string prefixed with '0x'."""
    if is_bytes(value):
        return add_0x_prefix(value.hex())
    else:
        raise DataTypeException("Data type should be bytes.")


def convert_hex_str_to_int(value: str) -> int:
    """Converts hex string prefixed with '0x' into int."""
    if is_str(value):
        return int(value, 16)
    else:
        raise DataTypeException("Data type should be string.")


def convert_hex_str_to_bytes(value: str) -> bytes:
    """Converts hex string prefixed with '0x' into bytes."""
    if is_str(value):
        return bytes.fromhex(remove_0x_prefix(value))
    else:
        raise DataTypeException("Data type should be string.")



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union, Tuple, Any, List, Optional

from iconsdk.builder.call_builder import Call
from iconsdk.builder.transaction_builder import Transaction
from iconsdk.exception import AddressException, DataTypeException
from iconsdk.providers.provider import Provider, MonitorSpec, Monitor
from iconsdk.signed_transaction import SignedTransaction
from iconsdk.utils import get_timestamp
from iconsdk.utils.convert_type import convert_int_to_hex_str
from iconsdk.utils.converter import convert, \
    get_block_template_to_convert_transactions_for_genesis
from iconsdk.utils.hexadecimal import add_0x_prefix, remove_0x_prefix
from iconsdk.utils.templates import BLOCK_0_1a, BLOCK_0_3, TRANSACTION_RESULT, TRANSACTION, BLOCK_0_1A_VERSION
from iconsdk.utils.validation import (
    is_block_height,
    is_hex_block_hash,
    is_predefined_block_value,
    is_score_address,
    is_wallet_address,
    is_T_HASH
)


class IconService:
    """
    The IconService class contains a set of API methods.
    It accepts a HTTPProvider which serves the purpose of
    connecting to HTTP and HTTPS based JSON-RPC servers.
    """
    DEFAULT_BLOCK_VERSION = BLOCK_0_1A_VERSION

    def __init__(self, provider: Provider):
        self.__provider = provider

    def get_block(self, value: Union[int, str], full_response: bool = False,
                  block_version: str = DEFAULT_BLOCK_VERSION) -> dict:
        """
        If param is height,
            1. Returns block information by block height
            2-1. Delegates to icx_getBlockByHeight RPC method (When Block Version is 0.1a)
            2-2. Delegates to icx_getBlock RPC method (When Block Version is 0.3)

        Or block hash,
            1. Returns block information by block hash
            2-1. Delegates to icx_getBlockByHash RPC method (When Block Version is 0.1a)
            2-2. Delegates to icx_getBlock RPC method (When Block Version is 0.3)

        Or string value same as `latest`,
            1. Returns the last block information
            2-1. Delegates to icx_getLastBlock RPC method (When Block Version is 0.1a)
            2-2. Delegates to icx_getBlock RPC method (When Block Version is 0.3)

        :param value:
            Integer of a block height
            or hash of a block prefixed with '0x'
            or `latest`
        :param full_response:
            Boolean to check whether get naive dict or refined data from server
        :param block_version:
            returning block format version

        :return result: Block data
        """

        # Nested method of returning right name of API method
        def return_infos_by_block_version(_prev_method: str) -> Tuple[str, Any, bool]:
            """ Returns API method name, block template, bool of full print by block version

            :param _prev_method: previous API methods.
                    For instance, icx_getBlockByHeight, icx_getBlockByHash and icx_getLastBlock
            :return: method name, block template, bool of full print
            """
            new_method = "icx_getBlock"
            if block_version == self.DEFAULT_BLOCK_VERSION:
                return _prev_method, BLOCK_0_1a, False
            else:
                return new_method, BLOCK_0_3, True

        # by height
        if is_block_height(value):
            params = {'height': add_0x_prefix(hex(value))}
            prev_method = 'icx_getBlockByHeight'
        # by hash
        elif is_hex_block_hash(value):
            params = {'hash': value}
            prev_method = 'icx_getBlockByHash'
        # last block
        elif is_predefined_block_value(value):
            params = None
            prev_method = 'icx_getLastBlock'
        else:
            raise DataTypeException("It's unrecognized block reference:{0!r}.".format(value))

        method, block_template, full_print = return_infos_by_block_version(prev_method)
        result = self.__provider.make_request(method, params, full_response)

        if not full_response:
            block_template = get_block_template_to_convert_transactions_for_genesis(result, block_template)
            result = convert(result, block_template, full_print)

        return result

    def get_total_supply(self, height: int = None, full_response: bool = False) -> Union[dict, int]:
        """
        Returns total ICX coin supply that has been issued
        Delegates to icx_getTotalSupply RPC method

        :param height: Block height
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: Total number of ICX coins issued
        """

        params = {}
        if height is not None:
            params["height"] = hex(height)

        result = self.__provider.make_request('icx_getTotalSupply', params, full_response=full_response)

        if full_response:
            return result
        else:
            return int(remove_0x_prefix(result), 16)

    def get_balance(self, address: str, height: int = None, full_response: bool = False) -> Union[dict, int]:
        """
        Returns the ICX balance of the given EOA or SCORE.
        Delegates to icx_getBalance RPC method.

        :param address: An address of EOA or SCORE. type(str)
        :param height: Block height. type(int)
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: Number of ICX coins
        """
        if is_score_address(address) or is_wallet_address(address):
            params = {'address': address}
            if height is not None:
                params["height"] = hex(height)

            result = self.__provider.make_request('icx_getBalance', params, full_response)

            if full_response:
                return result
            else:
                return int(remove_0x_prefix(result), 16)
        else:
            raise AddressException("Address is wrong.")

    def get_score_api(self, address: str, height: int = None, full_response: bool = False) -> Union[dict, list]:
        """
        Returns SCORE's external API list.
        Delegates to icx_getScoreApi RPC method.

        :param address: A SCORE address to be examined
        :param height: Block height
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: A list of API methods of the SCORE and its information
        """
        if not is_score_address(address):
            raise AddressException("SCORE Address is wrong.")

        params = {'address': address}
        if height is not None:
            params['height'] = hex(height)
        return self.__provider.make_request('icx_getScoreApi', params, full_response)

    def get_score_status(self, address: str, height: int = None, full_response: bool = False) -> dict:
        """
        Returns SCORE's status.
        Delegates to icx_getScoreStatus RPC method.

        :param address: A SCORE address to be examined
        :param height: Block height
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: Status information
        """
        if not is_score_address(address):
            raise AddressException("SCORE Address is wrong.")

        params = {'address': address}
        if height is not None:
            params['height'] = hex(height)
        return self.__provider.make_request('icx_getScoreStatus', params, full_response)

    def get_transaction_result(self, tx_hash: str, full_response: bool = False) -> dict:
        """
        Returns the transaction result requested by transaction hash.
        Delegates to icx_getTransactionResult RPC method.

        :param tx_hash: Hash of a transaction prefixed with '0x'
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return A transaction result object
        """
        if not is_T_HASH(tx_hash):
            raise DataTypeException("This hash value is unrecognized.")

        params = {'txHash': tx_hash}
        result = self.__provider.make_request('icx_getTransactionResult', params, full_response)

        if not full_response:
            result = convert(result, TRANSACTION_RESULT)

        return result

    def wait_transaction_result(self, tx_hash: str, full_response: bool = False) -> dict:
        """
        Returns the result of a transaction specified by the transaction hash like get_transaction_result,
        but waits for some time to get the transaction result instead of returning immediately
        if there is no finalized result.
        Delegates to icx_WaitTransactionResult RPC method.

        :param tx_hash: Hash of a transaction prefixed with '0x'
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return A transaction result object
        """
        if not is_T_HASH(tx_hash):
            raise DataTypeException("This hash value is unrecognized.")

        params = {'txHash': tx_hash}
        result = self.__provider.make_request('icx_waitTransactionResult', params, full_response)

        if not full_response:
            result = convert(result, TRANSACTION_RESULT)

        return result

    def get_transaction(self, tx_hash: str, full_response: bool = False) -> dict:
        """
        Returns the transaction information requested by transaction hash.
        Delegates to icx_getTransactionByHash RPC method.

        :param tx_hash: Transaction hash prefixed with '0x'
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: Information about a transaction
        """
        if not is_T_HASH(tx_hash):
            raise DataTypeException("This hash value is unrecognized.")

        params = {'txHash': tx_hash}
        result = self.__provider.make_request('icx_getTransactionByHash', params, full_response)

        if not full_response:
            result = convert(result, TRANSACTION)

        return result

    def call(self, call: object, full_response: bool = False) -> Union[dict, str]:
        """
        Calls SCORE's external function which is read-only without creating a transaction.
        Delegates to icx_call RPC method.

        :param call: Call object made by CallBuilder
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: Values returned by the executed SCORE function
        """
        if not isinstance(call, Call):
            raise DataTypeException("Call object is unrecognized.")

        params = {
            "to": call.to,
            "dataType": "call",
            "data": {
                "method": call.method
            }
        }

        if call.from_ is not None:
            params["from"] = call.from_

        if isinstance(call.params, dict):
            params["data"]["params"] = call.params

        if call.height is not None:
            params["height"] = call.height

        return self.__provider.make_request('icx_call', params, full_response)

    def send_transaction(self, signed_transaction: SignedTransaction, full_response: bool = False) -> Union[dict, str]:
        """
        Sends the transaction.
        Delegates to icx_sendTransaction RPC method.

        :param signed_transaction: A signed transaction object
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return: Transaction hash prefixed with '0x'
        """
        params = signed_transaction.signed_transaction_dict
        return self.__provider.make_request('icx_sendTransaction', params, full_response)

    def send_transaction_and_wait(self, signed_transaction: SignedTransaction, full_response: bool = False) -> dict:
        """
        Sends a transaction like icx_sendTransaction, then it will wait for the result of it for specified time.
        Delegates to icx_sendTransactionAndWait RPC method.

        :param signed_transaction: A signed transaction object
        :param full_response: Boolean to check whether get naive dict or refined data from server
        :return A transaction result object
        """
        params = signed_transaction.signed_transaction_dict
        result = self.__provider.make_request('icx_sendTransactionAndWait', params, full_response)

        if not full_response:
            result = convert(result, TRANSACTION_RESULT)

        return result

    def estimate_step(self, transaction: Transaction, full_response: bool = False) -> int:
        """
        Returns an estimated step of how much step is necessary to allow the transaction to complete.

        :param transaction: a raw transaction
        :param full_response: a boolean indicating whether or not it returns refined data
        :return: an estimated step
        """
        if not isinstance(transaction, Transaction):
            raise DataTypeException("Transaction object is unrecognized.")

        params = {
            "version": convert_int_to_hex_str(transaction.version) if transaction.version else "0x3",
            "from": transaction.from_,
            "to": transaction.to,
            "timestamp": convert_int_to_hex_str(transaction.timestamp) if transaction.timestamp else get_timestamp(),
            "nid": convert_int_to_hex_str(transaction.nid) if transaction.nid else "0x1"
        }

        if transaction.value is not None:
            params["value"] = convert_int_to_hex_str(transaction.value)

        if transaction.nonce is not None:
            params["nonce"] = convert_int_to_hex_str(transaction.nonce)

        if transaction.data_type is not None:
            params["dataType"] = transaction.data_type

        if transaction.data is not None:
            params["data"] = transaction.data

        result = self.__provider.make_request('debug_estimateStep', params, full_response)
        return result if full_response else int(result, 16)

    def get_trace(self, tx_hash: str) -> dict:
        """
        Get trace of the transaction

        :param tx_hash: Transaction hash prefixed with '0x'
        :return: trace
        """

        if not is_T_HASH(tx_hash):
            raise DataTypeException("This hash value is unrecognized.")

        params = {'txHash': tx_hash}
        result = self.__provider.make_request('debug_getTrace', params)
        return result

    def get_data_by_hash(self, _hash: str) -> str:
        """
        Returns data by hash.
        It can be used to retrieve data based on the hash algorithm (SHA3-256).
        Following data can be retrieved by a hash.
            - BlockHeader with the hash of the block
            - Validators with BlockHeader.NextValidatorsHash
            - Votes with BlockHeader.VotesHash
            - etc…
        Delegates to icx_getDataByHash RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp_extension.md#icx_getdatabyhash

        :param _hash: The hash value of the data to retrieve
        :return: A data object
        """
        params = {'hash': _hash}
        return self.__provider.make_request('icx_getDataByHash', params)

    def get_block_header_by_height(self, height: int) -> str:
        """
        Returns block header for specified height.
        Delegates to icx_getBlockHeaderByHeight RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp_extension.md#icx_getblockheaderbyheight

        :param height: The height of the block
        :return: A block header object
        """
        params = {'height': hex(height)}
        return self.__provider.make_request('icx_getBlockHeaderByHeight', params)

    def get_votes_by_height(self, height: int) -> str:
        """
        Returns votes for the block specified by height.
        Delegates to icx_getVotesByHeight RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp_extension.md#icx_getvotesbyheight

        :param height: The height of the block for votes
        :return: A votes object
        """
        params = {'height': hex(height)}
        return self.__provider.make_request('icx_getVotesByHeight', params)

    def get_proof_for_result(self, _hash: str, index: int) -> str:
        """
        Returns proof for the receipt. Proof, itself, may include the receipt.
        Delegates to icx_getProofForResult RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp_extension.md#icx_getproofforresult

        :param _hash: The hash value of the block including the result
        :param index: Index of the receipt in the block. 0 for the first
        :return: A proof object
        """
        params = {
            'hash': _hash,
            'index': hex(index),
        }
        return self.__provider.make_request('icx_getProofForResult', params)

    def get_proof_for_events(self, _hash: str, index: int, events: List[str] = []) -> dict:
        """
        Returns proof for the receipt and the events in it. The proof may include the data itself.
        Delegates to icx_getProofForEvents RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp_extension.md#icx_getproofforevents

        :param _hash: The hash value of the block including the result
        :param index: Index of the receipt in the block. 0 for the first
        :param events: List of indexes of the events in the receipt
        :return: A proof object
        """
        params = {
            'hash': _hash,
            'index': hex(index),
            'events': events,
        }
        return self.__provider.make_request('icx_getProofForEvents', params)

    def get_btp_network_info(self, id: int, height: int = None) -> dict:
        """
        Returns BTP Network information for specified height and ID.
        Delegates to btp_getNetworkInfo RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#btp_getNetworkInfo

        :param id: The id of the BTP network
        :param height: The height of the main block
        :return: A BTP Network information object
        """
        params = {'id': hex(id)}
        if height is not None:
            params['height'] = hex(height)
        return self.__provider.make_request('btp_getNetworkInfo', params)

    def get_btp_network_type_info(self, id: int, height: int = None) -> dict:
        """
        Returns BTP Network Type information for specified height and ID.
        Delegates to btp_getNetworkTypeInfo RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#btp_getNetworkTypeInfo

        :param id: The id of the BTP network type
        :param height: The height of the main block
        :return: A BTP Network Type information object
        """
        params = {'id': hex(id)}
        if height is not None:
            params['height'] = hex(height)
        return self.__provider.make_request('btp_getNetworkTypeInfo', params)

    def get_btp_messages(self, height: int, network_id: int) -> list:
        """
        Returns BTP messages for specified height and network ID.
        Delegates to btp_getMessages RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#btp_getMessages

        :param height: The height of the main block
        :param network_id: The id of the BTP network
        :return: A BTP Messages object
        """
        params = {'height': hex(height), 'networkID': hex(network_id)}
        return self.__provider.make_request('btp_getMessages', params)

    def get_btp_header(self, height: int, network_id: int) -> str:
        """
        Returns BTP block header for specified height and network ID.
        Delegates to btp_getHeader RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#btp_getHeader

        :param height: The height of the main block
        :param network_id: The id of the BTP network
        :return: A Base64 encoded BTP block header
        """
        params = {'height': hex(height), 'networkID': hex(network_id)}
        return self.__provider.make_request('btp_getHeader', params)

    def get_btp_proof(self, height: int, network_id: int) -> str:
        """
        Returns BTP block proof for specified height and network ID.
        Delegates to btp_getHeader RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#btp_getProof

        :param height: The height of the main block
        :param network_id: The id of the BTP network
        :return: A Base64 encoded BTP block proof
        """
        params = {'height': hex(height), 'networkID': hex(network_id)}
        return self.__provider.make_request('btp_getProof', params)

    def get_btp_source_information(self) -> dict:
        """
        Returns BTP source network information.
        Delegates to btp_getSourceInformation RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#btp_getSourceInformation

        :return: A BTP network information object
        """
        return self.__provider.make_request('btp_getSourceInformation')

    def monitor(self, spec: MonitorSpec, keep_alive: Optional[float] = None) -> Monitor:
        """
        Monitor events specified in the spec

        :param spec: Monitor specification
        :param keep_alive: interval to send keep-alive while it reads a message
        in fraction of seconds
        :return: Monitoring handle
        """
        return self.__provider.make_monitor(spec, keep_alive)

    def get_network_info(self) -> dict:
        """
        Returns basic network information
        Delegates to icx_getNetworkInfo RPC method.
        https://github.com/icon-project/goloop/blob/master/doc/jsonrpc_v3.md#icx_getNetworkInfo

        :return: Basic network information
        """
        return self.__provider.make_request('icx_getNetworkInfo')


#  Copyright 2023 ICON Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import List, Dict, Any

from iconsdk.providers.provider import MonitorSpec
from iconsdk.utils.typing.conversion import object_to_str


class EventFilter:
    def __init__(self, event: str, addr: str, indexed: int, *args):
        self.__event = event
        self.__addr = addr
        self.__indexed = list(args[0:indexed])
        self.__data = list(args[indexed:])

    def apply_to(self, obj: Dict[str, Any]):
        obj.update({
            "event": self.__event,
            "indexed": self.__indexed,
            "data": self.__data,
        })
        if self.__addr is not None:
            obj.update({
                "addr": self.__addr
            })

    def as_dict(self) -> Dict[str, Any]:
        obj = {}
        self.apply_to(obj)
        return obj


class EventMonitorSpec(MonitorSpec):
    def __init__(self, height: int | None,
                 filters: EventFilter | List[EventFilter],
                 logs: bool = False,
                 progress_interval: int = 0):
        self.__height = height
        self.__logs = logs
        if type(filters) is EventFilter:
            self.__filters = list([filters])
        else:
            self.__filters = filters
        self.__progress_interval = progress_interval

    def get_path(self) -> str:
        return 'event'

    def get_request(self) -> any:
        params = {}
        if self.__height is not None:
            params["height"] = self.__height
        if self.__logs:
            params["logs"] = True
        if self.__progress_interval > 0:
            params["progressInterval"] = self.__progress_interval
        if len(self.__filters) == 1:
            self.__filters[0].apply_to(params)
        else:
            params["filters"] = list(map(lambda a: a.as_dict(), self.__filters))
        return object_to_str(params)


class BlockMonitorSpec(MonitorSpec):
    def __init__(self, height: int | None,
                 filters: EventFilter|List[EventFilter]|None = None,
                 logs: bool = False):
        """
        Block monitoring spec

        :param height: Block height to start monitoring
        :param filters: One event filter or list of event filters
        :param logs: Whether the notification includes logs or not
        """
        self.__height = height
        self.__logs = logs
        if type(filters) is EventFilter:
            self.__filters = list([filters])
        else:
            self.__filters = filters

    def get_path(self) -> str:
        return 'block'

    def get_request(self) -> Dict[str, Any]:
        params = {}
        if self.__height is not None:
            params["height"] = self.__height
        if self.__logs:
            params["logs"] = True
        if len(self.__filters) == 1:
            self.__filters[0].apply_to(params)
        else:
            params["eventFilters"] = list(map(lambda a: a.as_dict(), self.__filters))
        return object_to_str(params)


class BTPMonitorSpec(MonitorSpec):
    def __init__(self, height: int | None,
                 network_id: int,
                 proof_flag: bool = False,
                 progress_interval: int = 0):
        """
        BTP Block event monitoring specification.
        https://github.com/icon-project/goloop/blob/master/doc/btp2_extension.md#block

        :param height: Height of the block to start monitoring
        :param network_id: Network ID of the BTP Network
        :param proof_flag: Whether it includes proof data or not
        :param progress_interval: Progress interval to notify progress
        """
        self.__height = height
        self.__network_id = network_id
        self.__proof_flag = proof_flag
        self.__progress_interval = progress_interval

    def get_path(self) -> str:
        return 'btp'

    def get_request(self) -> Dict[str, Any]:
        params = {
            "networkID": self.__network_id,
        }
        if self.__height is not None:
            params["height"] = self.__height
        if self.__proof_flag:
            params['proofFlag'] = True
        if self.__progress_interval > 0:
            params["progressInterval"] = self.__progress_interval
        return object_to_str(params)



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number
from typing import Mapping

bytes_types = (bytes, bytearray)
integer_types = (int,)
str_types = (str,)
string_types = (bytes, str, bytearray)


def is_integer(value) -> bool:
    return isinstance(value, integer_types) and not isinstance(value, bool)


def is_bytes(value) -> bool:
    return isinstance(value, bytes_types)


def is_str(value) -> bool:
    return isinstance(value, str_types)


def is_boolean(value) -> bool:
    return isinstance(value, bool)


def is_dict(obj) -> bool:
    return isinstance(obj, Mapping)


def is_list(obj) -> bool:
    return isinstance(obj, list)


def is_tuple(obj) -> bool:
    return isinstance(obj, tuple)


def is_null(obj) -> bool:
    return obj is None


def is_number(obj) -> bool:
    return isinstance(obj, Number)



# -*- coding: utf-8 -*-
# Copyright 2020 ICON Foundation Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

from ...wallet.wallet import Wallet

BASE_TYPES = {bool, bytes, int, str, Wallet}
TYPE_NAME_TO_TYPE = {_type.__name__: _type for _type in BASE_TYPES}


def is_base_type(value: type) -> bool:
    try:
        return value in BASE_TYPES
    except:
        return False


def name_to_type(type_name: str) -> type:
    return TYPE_NAME_TO_TYPE[type_name]


def isinstance_ex(value: Any, _type: type) -> bool:
    if not isinstance(value, _type):
        return False

    if type(value) is bool and _type is not bool:
        return False

    return True


def base_object_to_str(value: Any) -> str:
    if isinstance(value, Wallet):
        return value.get_address()
    elif isinstance(value, int):
        return hex(value)
    elif isinstance(value, bytes):
        return bytes_to_hex(value)
    elif isinstance(value, bool):
        return hex(value)
    elif isinstance(value, str):
        return value

    raise TypeError(f"Unsupported type: {type(value)}")


def object_to_str(value: Any) -> Any | None:
    if is_base_type(type(value)):
        return base_object_to_str(value)

    if isinstance(value, list):
        return [object_to_str(i) for i in value]

    if isinstance(value, dict):
        return {k: object_to_str(value[k]) for k in value}

    if value is None:
        return None

    raise TypeError(f"Unsupported type: {type(value)}")


def bytes_to_hex(value: bytes, prefix: str = "0x") -> str:
    return f"{prefix}{value.hex()}"



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module is reference to `hexadecimal`.
It is used for

- value starting with `0x`
- an address of a wallet starting with 'hx'
- an address of SCORE starting with 'cx'
"""
from re import match

from iconsdk.exception import DataTypeException
from iconsdk.utils.type import is_str


def is_0x_prefixed(value: str) -> bool:
    if not is_str(value):
        raise DataTypeException("Value type must be str. Got: {0}".format(repr(value)))
    return value.startswith('0x')


def remove_0x_prefix(value: str) -> str:
    if is_0x_prefixed(value):
        return value[2:]
    return value


def add_0x_prefix(value: str) -> str:
    if is_0x_prefixed(value):
        return value
    return '0x' + value


def is_hx_prefixed(value: str) -> bool:
    """Used for checking an address of a wallet."""
    if not is_str(value):
        raise DataTypeException("Value type must be str. Got: {0}.".format(repr(value)))
    return value.startswith('hx')


def remove_hx_prefix(value: str) -> str:
    if is_hx_prefixed(value):
        return value[2:]
    return value


def add_hx_prefix(value: str) -> str:
    if is_hx_prefixed(value):
        return value
    return 'hx' + value


def is_cx_prefixed(value: str) -> bool:
    """Used for checking an address of SCORE."""
    if not is_str(value):
        raise DataTypeException("Value type must be str. Got: {0}.".format(repr(value)))
    return value.startswith("cx")


def remove_cx_prefix(value: str) -> str:
    if is_cx_prefixed(value):
        return value[2:]
    return value


def add_cx_prefix(value: str) -> str:
    if is_cx_prefixed(value):
        return value
    return 'cx' + value


def is_lowercase_hex_string(value: str) -> bool:
    """Check whether value is hexadecimal format or not

    :param value: text
    :return: True(lowercase hexadecimal) otherwise False
    """
    try:
        result = None
        if isinstance(value, str):
            result = match('[0-9a-f]+', value)
        return result is not None and len(result.group(0)) == len(value)
    except Exception:
        raise DataTypeException("Not lowercase hex string. Get: {0}.".format(repr(value)))



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from os import path, walk
from zipfile import ZipFile, ZIP_DEFLATED

from iconsdk.exception import ZipException


def gen_deploy_data_content(_path: str) -> bytes:
    """Generate bytes of zip data of SCORE.

    :param _path: Path of the directory to be zipped.
    """
    if path.isdir(_path) is False and path.isfile(_path) is False:
        raise ValueError(f"Invalid path {_path}")
    try:
        memory_zip = InMemoryZip()
        memory_zip.zip_in_memory(_path)
    except ZipException:
        raise ZipException(f"Can't zip SCORE contents")
    else:
        return memory_zip.data


class InMemoryZip:
    """Class for compressing data in memory using zip and BytesIO."""

    def __init__(self):
        self._in_memory = BytesIO()

    @property
    def data(self) -> bytes:
        """Returns zip data

        :return: zip data
        """
        self._in_memory.seek(0)
        return self._in_memory.read()

    def zip_in_memory(self, _path: str):
        """Compress zip data (bytes) in memory.

        :param _path: The path of the directory to be zipped.
        """
        try:
            # when it is a zip file
            if path.isfile(_path):
                zf = ZipFile(_path, 'r', ZIP_DEFLATED, False)
                zf.testzip()
                with open(_path, mode='rb') as fp:
                    fp.seek(0)
                    self._in_memory.seek(0)
                    self._in_memory.write(fp.read())
            else:
                # root path for figuring out directory of tests
                tmp_root = None
                with ZipFile(self._in_memory, 'a', ZIP_DEFLATED, False, compresslevel=9) as zf:
                    for root, folders, files in walk(_path):
                        if 'package.json' in files:
                            tmp_root = root
                        if tmp_root and root.replace(tmp_root,'') == '/tests':
                            continue
                        if root.find('__pycache__') != -1:
                            continue
                        if root.find('/.') != -1:
                            continue
                        for file in files:
                            if file.startswith('.'):
                                continue
                            full_path = path.join(root, file)
                            zf.write(full_path)
        except ZipException:
            raise ZipException




from setuptools import setup, find_packages
from iconsdk.version import __version__

with open("README.md", 'r') as f:
    long_description = f.read()

with open("requirements.txt") as requirements:
    requires = list(requirements)

extras_requires = {
    'tests': ['pytest~=7.2.1']
}

setup(
    name='iconsdk',
    version=__version__,
    description='ICON SDK for Python is a collection of libraries which allow you to interact '
                'with a local or remote ICON node using an HTTP connection.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='ICON Foundation',
    author_email='foo@icon.foundation',
    url='https://github.com/icon-project/icon-sdk-python',
    packages=find_packages(exclude=['tests*']),
    install_requires=requires,
    extras_require=extras_requires,
    python_requires='>=3.7',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ]
)



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import traceback

from iconsdk.exception import DataTypeException
from iconsdk.utils.hexadecimal import remove_0x_prefix
from iconsdk.utils.templates import (ValueType, ConvertKeyName, RemoveKey, ExceptionHandle, BLOCK_0_1a,
                                     TRANSACTIONS_OF_GENESIS_BLOCK)
from iconsdk.utils.type import is_str


def convert(return_value: dict, return_template, full_print: bool = False) -> dict:
    """
    Convert return value in dict into predefined return template

    by
    - converting key name,
    - its own type
    - removing key.

    :param return_value: return value from ICON JSON-RPC API v3
    :param return_template: predefined return template
    :param full_print: bool
        True is returning all keys of block template with emtpy string value
        False is returning existing keys and value
    :return: converted return value in dict as converted by key and removed
    """
    if isinstance(return_template, dict):
        if "CHANGE" in return_template:
            return_value = _change_key(return_value, return_template["CHANGE"])
        if full_print is True:
            buf_return_value = dict()
            for key, value in return_template.items():
                if key == "CHANGE":
                    continue
                buf_return_value[key] = return_value[key] if key in return_value else ""
            return_value = buf_return_value

    if isinstance(return_value, dict) and isinstance(return_template, dict):
        buf_return_value = dict()
        for key, value in return_value.items():
            buf_return_value[key] = convert(value, return_template.get(key, None))
    elif isinstance(return_value, list) and isinstance(return_template, list):
        buf_return_value = list()
        for item in return_value:
            buf_return_value.append(convert(item, return_template[0]))
    elif isinstance(return_template, ValueType):
        buf_return_value = convert_value(return_value, return_template) if return_value != '' else return_value
    else:
        buf_return_value = return_value

    return buf_return_value


def _change_key(obj, change_dict):
    new_obj = copy.copy(obj)
    for key, change_value in change_dict.items():
        if isinstance(change_value, RemoveKey):
            if key in new_obj:
                del new_obj[key]
        elif isinstance(change_value, ConvertKeyName):
            if key in new_obj:
                del new_obj[key]
                new_obj[change_value.new_key] = obj[key]
        elif isinstance(change_value, ExceptionHandle):
            if key in new_obj:
                change_value.func(key, new_obj[key], new_obj)
        else:
            raise RuntimeError(f"Not expected change, {change_value}")

    return new_obj


def convert_value(value, value_type):
    if value == '':
        return value
    try:
        if value_type == ValueType.str:
            return _to_str(value)
        elif value_type == ValueType.int:
            return _to_int(value)
        elif value_type == ValueType.bytes:
            return _to_bytes(value)
        elif value_type == ValueType.hex_hash_number:
            return _to_hex_hash_number(value)
        elif value_type == ValueType.prefixed_hex_hash_number:
            return _to_prefixed_hex_hash_number(value)

    except BaseException as e:
        traceback.print_exc()
        logging.error(f"Error : {e}, value : {value_type}:{value}")

    return value


def _to_str(value):
    if isinstance(value, str):
        return value
    return str(value)


def _to_int(value):
    if isinstance(value, int):
        return value
    if isinstance(value, str):
        if value.startswith('0x') or value.startswith('-0x'):
            return int(value, 16)
        try:
            return int(value)
        except:
            pass
        return int(value, 16)


def _to_bytes(value: str) -> bytes:
    """Converts hex string prefixed with '0x' into bytes."""
    if is_str(value):
        return bytes.fromhex(remove_0x_prefix(value))
    else:
        raise DataTypeException("Data type should be string.")


def _to_hex_hash_number(value):
    if isinstance(value, int):
        return hex(value).split("0x")[1]
    if isinstance(value, str):
        if value.startswith('0x') or value.startswith('-0x'):
            return value.split("0x")[1]
        else:
            return value


def _to_prefixed_hex_hash_number(value):
    if isinstance(value, int):
        return hex(value)
    if isinstance(value, str):
        if value.startswith('0x') or value.startswith('-0x'):
            return value

        num = int(value, 16)
        hex(int(value, 16))
        if num >= 0:
            return '0x' + value
        else:
            return '-0x' + value


def get_block_template_to_convert_transactions_for_genesis(block, block_template):
    buf_block_template = copy.deepcopy(block_template)
    if block["height"] == 0:
        buf_block_template["confirmed_transaction_list" if block_template == BLOCK_0_1a else "transactions"] \
            = TRANSACTIONS_OF_GENESIS_BLOCK
    return buf_block_template
